/*  dcoexfunc.c is part of dcoex.

    dcoex is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    dcoex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dcoex; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*#include <string.h>*/
#include "glob.h"

/*--- DECLARATION OF globals (defined in glob.h) */
extern gsl_matrix *exprs1;
extern gsl_matrix *exprs2;
extern gsl_vector *rmeans1;
extern gsl_vector *rmeans2;

/*--- global scratch vectors...*/
extern gsl_vector *rind_c;
extern gsl_vector *cbase1_c;
extern gsl_vector *cbase2_c;
extern gsl_matrix *scratch1;
extern gsl_matrix *scratch2;

extern gsl_rng *r_perm;


Pattern*
ptrn_alloc(int npt)
{
  Pattern *vec; /* pointer to be returned */
  int i;

  /*--- alloc vector of Pattern structs */
  vec = (Pattern*) malloc(npt*sizeof(Pattern));

    /*--- alloc the gsl_vectors each bc points to */
  for(i=0; i<npt; i++)
    {
      (vec+i)->rind   = gsl_vector_calloc(exprs1->size1);
      (vec+i)->cbase1 = gsl_vector_calloc(exprs1->size2);
      (vec+i)->cbase2 = gsl_vector_calloc(exprs2->size2);
    }
  return(vec);
}

/*-----------------------------------------------------------------------------------

rand_bin_gen

generates randomly 0 or 1
---------------------------------------------------------------------------------- */

int
rand_bin_gen(double chance)
{
  /*int oldseed; */
  const gsl_rng_type * T;
  gsl_rng * r;
  double u;

  gsl_rng_default_seed=gsl_rng_default_seed + rand();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);
  u = gsl_rng_uniform (r);

  if (u<chance) u = 1;

  else u = 0;

  gsl_rng_free (r);

  return u;

}

/*-----------------------------------------------------------------------------------

  ptrn_free

  description:
  frees  vector of patterns
---------------------------------------------------------------------------------- */

void
ptrn_free(Pattern *vec, int npt)
{
  int i;

  /*--- free gsl-vectors */
  for(i=0; i<npt; i++)
    {
      gsl_vector_free((vec+i)->rind);
      gsl_vector_free((vec+i)->cbase1);
      gsl_vector_free((vec+i)->cbase2);
    }

  /* free ptrn structs */
  free(vec);
}

/*-----------------------------------------------------------------------------------

  ptrn_rind

  description:
  case -1:
  takes allocated ptrn-vector and randomly initializes the row indices
  default:
  takes given matrix and checks if valid, else case -1

---------------------------------------------------------------------------------- */

int
ptrn_rind(Pattern *ptvec, int count, double rprob, int rows, gsl_matrix * smat, int have_sfile, double chance)
{
  int i,
      test;

  switch(have_sfile)
  {
    case -1:   /*--- init rows */
               for(i=0; i<(ptvec)->rind->size; i++)
               {
                  test=rand_bin_gen(chance);
                  gsl_vector_set((ptvec)->rind , i , test);
               }
               break;
    default:
	       for(i=0; i<(ptvec)->rind->size; i++)
	       {
                  gsl_vector_set((ptvec)->rind , i ,(gsl_matrix_get(smat, i, (count-1))));
               }

               break;
  }

  return(0);
}

/*-----------------------------------------------------------------------------------

  ptrn_val

  description:
  calculates some values of a pattern;
---------------------------------------------------------------------------------- */

int
ptrn_val (Pattern * ptrn)
{

  double tmp;
  int i;
  /*--- calculate number of rows */
  ptrn->nr =  gsl_blas_dasum (ptrn->rind);

  /*--- calculate column bases */
  if (ptrn->nr == 0)
    {
      printf ("\n ERROR: ptrn_val: \t zero row or column indices\n");
      return (-1);
    }

  gsl_blas_dgemv  (CblasTrans, 1., exprs1, ptrn->rind, 0., ptrn->cbase1);
  gsl_blas_dgemv  (CblasTrans, 1., exprs2, ptrn->rind, 0., ptrn->cbase2);
  gsl_vector_scale(ptrn->cbase1, 1./ ptrn->nr);
  gsl_vector_scale(ptrn->cbase2, 1./ ptrn->nr);

  /*--- calculate ptrn-means */
  tmp = 0;
  for(i=0; i<ptrn->cbase1->size; i++)
    {
      tmp += gsl_vector_get(ptrn->cbase1,i);
    }
  ptrn->mean1 = tmp/ptrn->cbase1->size;

  tmp = 0;

  for(i=0; i<ptrn->cbase2->size; i++)
    {
      tmp += gsl_vector_get(ptrn->cbase2,i);
    }

  ptrn->mean2 = tmp/ptrn->cbase2->size;

  /*--- calculate ptrn-res */
  ptrn->res = ptrn_res(ptrn);

  /*--- calculate ptrn-volume*/
  ptrn->vol = (int) ptrn->cbase1->size * ptrn->nr;

  return 0;
}

/*======================================================================================*/
/* CONCERNING ACTIONS                                                                   */
/*======================================================================================*/

/*-----------------------------------------------------------------------------------

  ac_valnr1

  description:
  determines if row/col is to be added or removed
  from a pattern and sets the appropriate ac-values
---------------------------------------------------------------------------------- */

int
ac_val(Action *ac, Pattern *ptrn, int ind, double alpha)
{

  /*--- copy passed values to appropriate places*/
  ac->ind = ind;
  ac->ptrn = (Pattern*) ptrn;

  /*--- set the arflag */
  if(gsl_vector_get(ptrn->rind,ind)==0)
    { /* row not allready in ptrn */
      ac->arflag = 1;
    }
  else if(gsl_vector_get(ptrn->rind,ind)==1)
    { /* row is already there */
      ac->arflag = 0;
    }
  else
    {
      printf("\n moi_ac_det: improper rind-value!\n");
      return(-1);
    }

  /*--- calculate the gain */
  ac->gain = ac_gain( ptrn, ac, alpha);

  return(0);
}

/*-----------------------------------------------------------------------------------

  ac_count

  description:
  count positive gains in acvec
---------------------------------------------------------------------------------- */

int
ac_count(const Action *ac, int nac)
{
  int i, cnt=0;

  for(i=0; i<nac; i++)
  {
    if((ac+i)->gain > 0)
    {
      cnt++;
    }
  }
  return(cnt);
}

/*-----------------------------------------------------------------------------------

  ac_sort

  description:
  sorts nac actions following a pointer
---------------------------------------------------------------------------------- */

void
ac_sort(Action *ac, int nac)
{

  qsort ((void*) ac, nac, sizeof(Action), ac_comp);

}

/*-----------------------------------------------------------------------------------

  ac_comp

  description:
  comparison function for qsort working on actions
---------------------------------------------------------------------------------- */

int
ac_comp(const void *a, const void *b)
{

  const Action  *aa = (const Action *) a;
  const Action  *ab = (const Action *) b;

  return (aa->gain > ab->gain) - (aa->gain < ab->gain);

}

/*-----------------------------------------------------------------------------------

  ac_swap

  description:
  swaps its to arguments
---------------------------------------------------------------------------------- */

void
ac_swap(Action *a1,Action *a2)
{
  size_t tmp[sizeof(Action)];

  memcpy(tmp, a1 , sizeof(Action));
  memcpy(a1 , a2 , sizeof(Action));
  memcpy(a2 , tmp, sizeof(Action));
}


/*-----------------------------------------------------------------------------------

  ac_rwo

  description:
  changes nac actions to the random weight order
---------------------------------------------------------------------------------- */

int
ac_rwo(Action *ac, int nac, int times)
{

  gsl_rng *rgen; /* pointer to random number generator */
  int i,e1,e2;
  double p,r;

  rgen = r_perm;

  if(nac < 2) return(0);


  for(i=0; i<times; i++)
  {
    /*--- sort actions first */
    ac_sort(ac, nac);
    r = (ac+nac-1)->gain - ac->gain;

    e1 = -2E5;
    e2 = -2E5;
    /*--- choose two different random elements */
    while(e1==e2)
    {
      e1 = gsl_rng_uniform_int(rgen, nac);
      e2 = gsl_rng_uniform_int(rgen, nac);
    }
    /*--- check if they are to be swapped */
    p = 0.5 + ((ac+e2)->gain - (ac+e1)->gain)/(2*r);
    if(gsl_rng_uniform(rgen) < p)
    {
      ac_swap(ac+e1, ac+e2);
    }
  }

  return(0);
}


/*-----------------------------------------------------------------------------------

  ac_perf

  description:
  perform actions pointing to the SAME PATTERN
---------------------------------------------------------------------------------- */

int
ac_perf(const Action *ac, int nac)
{

  double tmp, B1,B2;
  int i, ptnr;
  Pattern *ptrn;


  ptrn = ac->ptrn;
  ptnr = ptrn->nr;

  /* calculate new row indices an nr of ptrn; */
  B1 = 0;
  B2 = 0;
  for(i=0; i<nac; i++)
    {
      if((ac+i)->arflag == 0)
        {
          ptrn->nr -= 1;
          gsl_vector_set(ptrn->rind,(ac+i)->ind,0);
        }
      else if((ac+i)->arflag == 1)
        {
          ptrn->nr +=1;
          gsl_vector_set(ptrn->rind,(ac+i)->ind,1);
        }
      B1 += (ac+i)->B1;
      B2 += (ac+i)->B2;
    }

  B1 = B1/(ptrn->rind->size -1);
  B2 = B2/(ptrn->rind->size -1);

  /*--- calculate new col bases (indices can  have changed quite a bit) */
  gsl_blas_dgemv  (CblasTrans, 1., exprs1, ptrn->rind, 0., ptrn->cbase1);
  gsl_blas_dgemv  (CblasTrans, 1., exprs2, ptrn->rind, 0., ptrn->cbase2);
  gsl_vector_scale(ptrn->cbase1, 1./ ptrn->nr);
  gsl_vector_scale(ptrn->cbase2, 1./ ptrn->nr);

  /*--- calculate new ptrn-mean ...  */
  tmp = 0;
  for(i=0; i<ptrn->cbase1->size; i++)
    {
      tmp += gsl_vector_get(ptrn->cbase1,i);
    }
  ptrn->mean1 = tmp/ptrn->cbase1->size;

  tmp = 0;
  for(i=0; i<ptrn->cbase2->size; i++)
    {
      tmp += gsl_vector_get(ptrn->cbase2,i);
    }
  ptrn->mean2 = tmp/ptrn->cbase2->size;


  /*--- calculate new ptrn-res */
  ptrn->res = ptrn_res (ptrn);            /* new res */
  ptrn->vol = ptrn->nr * ptrn->cbase1->size;    /* new vol */

  return(0);
}

/*======================================================================================*/
/* CONCERNING MISC                                                                      */
/*======================================================================================*/

/*-----------------------------------------------------------------------------------

  matrix_perm

  description:
  shuffles expression values gene wise []
---------------------------------------------------------------------------------- */

int
matrix_perm( gsl_matrix *mat1, gsl_matrix *mat2)
{
  int i;
  gsl_vector_view row1, row2;
  gsl_permutation * p1 = gsl_permutation_calloc(mat1->size2); /*-- ident. perm */
  gsl_permutation * p2 = gsl_permutation_calloc(mat2->size2); /*-- ident. perm */

  for(i=0; i<mat1->size1; i++)
    {
      gsl_ran_shuffle(r_perm, p1->data, mat1->size2, sizeof(size_t)); /*-- permute */
      gsl_ran_shuffle(r_perm, p2->data, mat2->size2, sizeof(size_t)); /*-- permute */

      row1 = gsl_matrix_row(mat1,i); /*-- ith row of mat1 */
      row2 = gsl_matrix_row(mat2,i); /*-- ith row of mat2 */
      gsl_permute_vector(p1, &row1.vector);
      gsl_permute_vector(p2, &row2.vector);
    }

  gsl_permutation_free(p1);
  gsl_permutation_free(p2);
  return(0);
}


/*-----------------------------------------------------------------------------------

  matrix_scale

  description:
  scales matrix row-wise (subtract mean and divide by sd)
---------------------------------------------------------------------------------- */

int matrix_scale(gsl_matrix *mat)
{

  gsl_vector_view row;
  gsl_vector *ones, *rmeans;
  double dp;
  int i;

  rmeans = gsl_vector_alloc(mat->size1);
  ones   = gsl_vector_alloc(mat->size2);

  /*--- calc row means */
  gsl_vector_set_all(ones, 1.);
  gsl_blas_dgemv(CblasNoTrans, 1., mat, ones, 0., rmeans);
  gsl_vector_scale(rmeans, 1/(double)mat->size2);

  /*--- scale matrix */
  for(i=0; i<mat->size1; i++)
    {
      row = gsl_matrix_row(mat,i);
      gsl_vector_set_all(ones, 1.);
      gsl_vector_scale(ones,gsl_vector_get(rmeans,i));
      gsl_vector_sub(ones,&row.vector);
      gsl_blas_ddot(ones,ones,&dp);          /* ssq           */
      dp = sqrt(dp/(mat->size2 -1));         /* sd            */
      gsl_vector_scale(&row.vector,(1/dp));  /* scale the row */
    }

  gsl_vector_free(ones);
  gsl_vector_free(rmeans);

  return(0);
}


/*======================================================================================*/
/* SECONDARY PROGS                                                                      */
/*======================================================================================*/

/*-----------------------------------------------------------------------------------

  ptrn_res

  description:
  calculates residual of a pattern; row and col means have to be there
---------------------------------------------------------------------------------- */

double
ptrn_res (Pattern *ptrn)
{

  double r1, r2;

  /*--- calculate fit for grp 1*/
  vec_expand (rmeans1, ptrn->cbase1, scratch1); /* fit_ij = rb_i + cb_j ; null ausserhalb der indexmengen */
  gsl_matrix_add_constant (scratch1, -ptrn->mean1);

  /*--- calculater mean squared res sum for grp 1*/
  gsl_matrix_scale (scratch1, (-1.));
  gsl_matrix_add (scratch1, exprs1);
  gsl_matrix_mul_elements (scratch1, scratch1); /* fit_ij = residual_ij**2  ; --nicht-- null ausserhalb der indexmengen */
  matrix_zero (scratch1, ptrn->rind);
  r1 = matrix_sum (scratch1) / ((ptrn->cbase1->size -1) * (ptrn->nr-1));

  /*--- calculate fit for grp 2*/
  vec_expand (rmeans2, ptrn->cbase2, scratch2); /* fit_ij = rb_i + cb_j ; null ausserhalb der indexmengen */
  gsl_matrix_add_constant (scratch2, -ptrn->mean2);

  /*--- calculater mean squared res sum for grp 1*/
  gsl_matrix_scale (scratch2, (-1.));
  gsl_matrix_add (scratch2, exprs2);
  gsl_matrix_mul_elements (scratch2, scratch2); /* fit_ij = residual_ij**2  ; --nicht-- null ausserhalb der indexmengen */
  matrix_zero (scratch2, ptrn->rind);
  r2 = matrix_sum (scratch2) / ((ptrn->cbase2->size -1) * (ptrn->nr -1));

  ptrn->res1 = r1;
  ptrn->res2 = r2;

  return (r1/r2);

}


/* ----------------------------------------------------------------------------------

  vec_expand

  description:
  takes two gsl_vectors a and b and returns a pointer to a gsl_matrix
  corresponding to  res_ij = a_j + b_j

  important:
  res has to be  allocated to have length(a) rows and length(b) cols !
---------------------------------------------------------------------------------- */

void
vec_expand (const gsl_vector * a, const gsl_vector * b, gsl_matrix * res)
{

  int as = a->size;
  int i;
  gsl_vector_view row;

  gsl_matrix_set_zero (res);    /* to make sure */
  for (i = 0; i < as; i++)
    {
      row = gsl_matrix_row (res, i);                                    /* ith row       */
      gsl_vector_add (&row.vector, b);                                  /* equal it to b */
      gsl_vector_add_constant (&row.vector, gsl_vector_get (a, i));     /* add a_i       */
    }
}



/*-----------------------------------------------------------------------------------

  matrix_sum

  description:
  returns sum of elements of a gsl_matrix
---------------------------------------------------------------------------------- */

double
matrix_sum (const gsl_matrix * mat)
{

  int i, j, nr, nc;
  double res;

  nr = mat->size1;
  nc = mat->size2;

  res = 0;
  for (i = 0; i < nr; i++)
    {
      for (j = 0; j < nc; j++)
        {
          res += gsl_matrix_get (mat, i, j);
        }
    }

  return res;
}


/*-----------------------------------------------------------------------------------

  matrix_zero

  description:
  zeros elements of matrix according to row / column indices

---------------------------------------------------------------------------------- */

void
matrix_zero (gsl_matrix *mat, const gsl_vector *rind)
{

  gsl_vector_view col;
  unsigned i;

  for (i = 0; i<mat->size2; i++)
    {
      col = gsl_matrix_column(mat, i);
      gsl_vector_mul (&col.vector, rind); /* in every col zero the not included rows */
    }
}











/*-----------------------------------------------------------------------------------

  ac_gain

  description:
  calculates gain of a action in the supervised setting.
  we copy the old pattern vals and use the bc_res function
---------------------------------------------------------------------------------- */

double
ac_gain(Pattern * bc,  Action * ac, double alpha)
{

  double r1, r2,  gain,v1,v2,B1,B2,A1,A2;
  int vol;
  gsl_vector_view row1, row2;

  row1 = gsl_matrix_row(exprs1, ac->ind);
  row2 = gsl_matrix_row(exprs2, ac->ind);

  gsl_vector_memcpy(cbase1_c,bc->cbase1);
  gsl_vector_memcpy(cbase2_c,bc->cbase2);
  gsl_vector_scale(cbase1_c, -1.);
  gsl_vector_scale(cbase2_c, -1.);

  gsl_vector_add(cbase1_c,&row1.vector);
  gsl_vector_add(cbase2_c,&row2.vector);

  gsl_vector_add_constant(cbase1_c, -gsl_vector_get(rmeans1,ac->ind) + bc->mean1);
  gsl_vector_add_constant(cbase2_c, -gsl_vector_get(rmeans2,ac->ind) + bc->mean2);

  gsl_blas_ddot(cbase1_c, cbase1_c, &r1);
  gsl_blas_ddot(cbase2_c, cbase2_c, &r2);

  r1 /= (double)(cbase1_c->size -1);
  r2 /= (double)(cbase2_c->size -1);

  B1 = r1/((double)(1.));
  B2 = r2/((double)(1.));

  A1 = bc->res1 - B1;
  A2 = bc->res2 - B2;

  ac->del1 = bc->res1/bc->res2*A2/A1;  /* effect without refitting  */
  ac->B1   = B1;
  ac->B2   = B2;

  if(ac->arflag==1)
    {
      vol =  bc->vol + cbase1_c->size;
      v1 =  1.*(vol - bc->vol)/bc->vol;
      v2 =  1.*(vol - bc->vol)/bc->vol;

      gain =  alpha * (bc->res1/bc->res2 - r1/r2) + (1.-alpha)*v1 ;
    }
  else
    {
      vol =  bc->vol - cbase1_c->size;
      v1 =  1.*(vol - bc->vol)/bc->vol;
      v2 =  1.*(vol - bc->vol)/bc->vol;

      gain = alpha * (r1/r2 - bc->res1/bc->res2) + (1. -alpha)*v1;
    }

  return(gain);
}






/*-----------------------------------------------------------------------------------

  moi_matrix_print

  description:
  print a gsl_matrix to screen
---------------------------------------------------------------------------------- */

int
moi_matrix_print (gsl_matrix * mat)
{
  int i, j;

  for (i = 0; i < (int)mat->size1; i++)
    {
      for (j = 0; j < (int)mat->size2; j++)
        {
          printf (" %1.2f ", gsl_matrix_get (mat, i, j));
        }
      printf ("\n");
    }

  return (0);
}


/*-----------------------------------------------------------------------------------

  moi_bc_print

  description:
  print a bc to screen
---------------------------------------------------------------------------------- */

int
moi_bc_print (Pattern * bc)
{

  printf ("\n\n");
  printf ("bc_res  :  \t%e\n", bc->res);
  printf ("bc_vol  :\t\t%d\n", bc->vol);
  printf ("bc_mean1:\t\t%f\n", bc->mean1);
  printf ("bc_mean2:\t\t%f\n", bc->mean2);
  printf ("#rows   : \t\t%d\n", bc->nr);

  /* ... to be continued ... */

  return (0);
}


